# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest
from datetime import datetime

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib, assertRaisesRegex
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import NodeImage
from libcloud.test.secrets import DIGITALOCEAN_v1_PARAMS, DIGITALOCEAN_v2_PARAMS
from libcloud.utils.iso8601 import UTC
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.common.digitalocean import DigitalOcean_v1_Error
from libcloud.compute.drivers.digitalocean import DigitalOceanNodeDriver

try:
    import simplejson as json
except ImportError:
    import json  # NOQA


# class DigitalOceanTests(unittest.TestCase, TestCaseMixin):
class DigitalOcean_v2_Tests(LibcloudTestCase):
    def setUp(self):
        DigitalOceanNodeDriver.connectionCls.conn_class = DigitalOceanMockHttp
        DigitalOceanMockHttp.type = None
        self.driver = DigitalOceanNodeDriver(*DIGITALOCEAN_v2_PARAMS)

    def test_v1_Error(self):
        self.assertRaises(
            DigitalOcean_v1_Error,
            DigitalOceanNodeDriver,
            *DIGITALOCEAN_v1_PARAMS,
            api_version="v1",
        )

    def test_v2_uses_v1_key(self):
        self.assertRaises(
            InvalidCredsError,
            DigitalOceanNodeDriver,
            *DIGITALOCEAN_v1_PARAMS,
            api_version="v2",
        )

    def test_authentication(self):
        DigitalOceanMockHttp.type = "UNAUTHORIZED"
        self.assertRaises(InvalidCredsError, self.driver.list_nodes)

    def test_list_images_success(self):
        images = self.driver.list_images()
        self.assertTrue(len(images) >= 1)

        image = images[0]
        self.assertTrue(image.id is not None)
        self.assertTrue(image.name is not None)

    def test_list_sizes_success(self):
        sizes = self.driver.list_sizes()
        self.assertTrue(len(sizes) >= 1)

        size = sizes[0]
        self.assertTrue(size.id is not None)
        self.assertEqual(size.name, "512mb")
        self.assertEqual(size.ram, 512)

        size = sizes[1]
        self.assertTrue(size.id is not None)
        self.assertEqual(size.name, "1gb")
        self.assertEqual(size.ram, 1024)

    def test_list_sizes_filter_by_location_success(self):
        location = self.driver.list_locations()[1]
        sizes = self.driver.list_sizes(location=location)
        self.assertTrue(len(sizes) >= 1)

        size = sizes[0]
        self.assertTrue(size.id is not None)
        self.assertEqual(size.name, "512mb")
        self.assertTrue(location.id in size.extra["regions"])

        location = self.driver.list_locations()[1]
        location.id = "doesntexist"
        sizes = self.driver.list_sizes(location=location)
        self.assertEqual(len(sizes), 0)

    def test_list_locations_success(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) == 2)

        location = locations[0]
        self.assertEqual(location.id, "nyc1")
        self.assertEqual(location.name, "New York 1")

        locations = self.driver.list_locations(ex_available=True)
        self.assertTrue(len(locations) == 2)

        locations = self.driver.list_locations(ex_available=False)
        self.assertTrue(len(locations) == 3)

    def test_list_nodes_success(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].name, "ubuntu-s-1vcpu-1gb-sfo3-01")
        self.assertEqual(nodes[0].public_ips, ["128.199.13.158"])
        self.assertEqual(nodes[0].extra["image"]["id"], 69463186)
        self.assertEqual(nodes[0].extra["size_slug"], "s-1vcpu-1gb")
        self.assertEqual(len(nodes[0].extra["tags"]), 0)

    def test_list_nodes_fills_created_datetime(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(nodes[0].created_at, datetime(2020, 10, 15, 13, 58, 22, tzinfo=UTC))

    def test_create_node_invalid_size(self):
        image = NodeImage(id="invalid", name=None, driver=self.driver)
        size = self.driver.list_sizes()[0]
        location = self.driver.list_locations()[0]

        DigitalOceanMockHttp.type = "INVALID_IMAGE"
        expected_msg = (
            r"You specified an invalid image for Droplet creation."
            + r" \(code: (404|HTTPStatus.NOT_FOUND)\)"
        )
        assertRaisesRegex(
            self,
            Exception,
            expected_msg,
            self.driver.create_node,
            name="test",
            size=size,
            image=image,
            location=location,
        )

    def test_reboot_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "REBOOT"
        result = self.driver.reboot_node(node)
        self.assertTrue(result)

    def test_create_image_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "SNAPSHOT"
        result = self.driver.create_image(node, "My snapshot")
        self.assertTrue(result)

    def test_get_image_success(self):
        image = self.driver.get_image(12345)
        self.assertEqual(image.name, "My snapshot")
        self.assertEqual(image.id, "12345")
        self.assertEqual(image.extra["distribution"], "Ubuntu")

    def test_delete_image_success(self):
        image = self.driver.get_image(12345)
        DigitalOceanMockHttp.type = "DESTROY"
        result = self.driver.delete_image(image)
        self.assertTrue(result)

    def test_ex_power_on_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "POWERON"
        result = self.driver.ex_power_on_node(node)
        self.assertTrue(result)

    def test_ex_shutdown_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "SHUTDOWN"
        result = self.driver.ex_shutdown_node(node)
        self.assertTrue(result)

    def test_ex_hard_reboot_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "POWERCYCLE"
        result = self.driver.ex_hard_reboot(node)
        self.assertTrue(result)

    def test_ex_rebuild_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "REBUILD"
        result = self.driver.ex_rebuild_node(node)
        self.assertTrue(result)

    def test_ex_resize_node_success(self):
        node = self.driver.list_nodes()[0]
        size = self.driver.list_sizes()[0]
        DigitalOceanMockHttp.type = "RESIZE"
        result = self.driver.ex_resize_node(node, size)
        self.assertTrue(result)

    def test_destroy_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "DESTROY"
        result = self.driver.destroy_node(node)
        self.assertTrue(result)

    def test_ex_change_kernel_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "KERNELCHANGE"
        result = self.driver.ex_change_kernel(node, 7515)
        self.assertTrue(result)

    def test_ex_enable_ipv6_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "ENABLEIPV6"
        result = self.driver.ex_enable_ipv6(node)
        self.assertTrue(result)

    def test_ex_rename_node_success(self):
        node = self.driver.list_nodes()[0]
        DigitalOceanMockHttp.type = "RENAME"
        result = self.driver.ex_rename_node(node, "fedora helios")
        self.assertTrue(result)

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()
        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].extra["id"], 7717)
        self.assertEqual(keys[0].name, "test1")
        self.assertEqual(keys[0].public_key, "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAAAQQDGk5 example")

    def test_create_key_pair(self):
        DigitalOceanMockHttp.type = "CREATE"
        key = self.driver.create_key_pair(
            name="test1", public_key="ssh-rsa AAAAB3NzaC1yc2EAAAADAQsxRiUKn example"
        )
        self.assertEqual(key.name, "test1")
        self.assertEqual(key.fingerprint, "f5:d1:78:ed:28:72:5f:e1:ac:94:fd:1f:e0:a3:48:6d")

    def test_delete_key_pair(self):
        key = self.driver.list_key_pairs()[0]
        result = self.driver.delete_key_pair(key)
        self.assertTrue(result)

    def test__paginated_request_single_page(self):
        nodes = self.driver._paginated_request("/v2/droplets", "droplets")
        self.assertEqual(nodes[0]["name"], "ubuntu-s-1vcpu-1gb-sfo3-01")
        self.assertEqual(nodes[0]["image"]["id"], 69463186)
        self.assertEqual(nodes[0]["size_slug"], "s-1vcpu-1gb")

    def test__paginated_request_two_pages(self):
        DigitalOceanMockHttp.type = "PAGE_ONE"
        nodes = self.driver._paginated_request("/v2/droplets", "droplets")
        self.assertEqual(len(nodes), 2)

    def test_list_volumes(self):
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 1)
        volume = volumes[0]
        self.assertEqual(volume.id, "62766883-2c28-11e6-b8e6-000f53306ae1")
        self.assertEqual(volume.name, "example")
        self.assertEqual(volume.size, 4)
        self.assertEqual(volume.driver, self.driver)

    def test_list_volumes_empty(self):
        DigitalOceanMockHttp.type = "EMPTY"
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 0)

    def test_create_volume(self):
        nyc1 = [r for r in self.driver.list_locations() if r.id == "nyc1"][0]
        DigitalOceanMockHttp.type = "CREATE"
        volume = self.driver.create_volume(4, "example", nyc1)
        self.assertEqual(volume.id, "62766883-2c28-11e6-b8e6-000f53306ae1")
        self.assertEqual(volume.name, "example")
        self.assertEqual(volume.size, 4)
        self.assertEqual(volume.driver, self.driver)

    def test_attach_volume(self):
        node = self.driver.list_nodes()[0]
        volume = self.driver.list_volumes()[0]
        DigitalOceanMockHttp.type = "ATTACH"
        resp = self.driver.attach_volume(node, volume)
        self.assertTrue(resp)

    def test_detach_volume(self):
        volume = self.driver.list_volumes()[0]
        DigitalOceanMockHttp.type = "DETACH"
        resp = self.driver.detach_volume(volume)
        self.assertTrue(resp)

    def test_destroy_volume(self):
        volume = self.driver.list_volumes()[0]
        DigitalOceanMockHttp.type = "DESTROY"
        resp = self.driver.destroy_volume(volume)
        self.assertTrue(resp)

    def test_list_volume_snapshots(self):
        volume = self.driver.list_volumes()[0]
        snapshots = self.driver.list_volume_snapshots(volume)
        self.assertEqual(len(snapshots), 3)
        snapshot1, snapshot2, snapshot3 = snapshots
        self.assertEqual(snapshot1.id, "c0def940-9324-11e6-9a56-000f533176b1")
        self.assertEqual(snapshot2.id, "c2036724-9343-11e6-aef4-000f53315a41")
        self.assertEqual(snapshot3.id, "d347e033-9343-11e6-9a56-000f533176b1")

    def test_create_volume_snapshot(self):
        volume = self.driver.list_volumes()[0]
        DigitalOceanMockHttp.type = "CREATE"
        snapshot = self.driver.create_volume_snapshot(volume, "test-snapshot")
        self.assertEqual(snapshot.id, "c0def940-9324-11e6-9a56-000f533176b1")
        self.assertEqual(snapshot.name, "test-snapshot")
        self.assertEqual(volume.driver, self.driver)

    def test_delete_volume_snapshot(self):
        volume = self.driver.list_volumes()[0]
        snapshot = self.driver.list_volume_snapshots(volume)[0]
        DigitalOceanMockHttp.type = "DELETE"
        result = self.driver.delete_volume_snapshot(snapshot)
        self.assertTrue(result)

    def test_ex_get_node_details(self):
        node = self.driver.ex_get_node_details("3164444")
        self.assertEqual(node.name, "example.com")
        self.assertEqual(node.public_ips, ["36.123.0.123"])
        self.assertEqual(node.extra["image"]["id"], 12089443)
        self.assertEqual(node.extra["size_slug"], "8gb")
        self.assertEqual(len(node.extra["tags"]), 2)

    def test_ex_create_floating_ip(self):
        nyc1 = [r for r in self.driver.list_locations() if r.id == "nyc1"][0]
        floating_ip = self.driver.ex_create_floating_ip(nyc1)

        # Note that this is the ID. There is no real ID for a floating IP at
        # DigitalOcean, but the IP is unique so we can use that instead.
        self.assertEqual(floating_ip.id, "167.138.123.111")
        self.assertEqual(floating_ip.ip_address, "167.138.123.111")
        self.assertEqual(floating_ip.extra["region"]["slug"], "nyc1")
        # The newly created floating IP reserved to a region is not
        # associated with any droplet. See the DigitalOcean API docs
        # how to create a floating IP that is associated with an instance
        # from the start. This API call creates an unattached IP.
        self.assertIsNone(floating_ip.node_id)

    def test_ex_delete_floating_ip(self):
        nyc1 = [r for r in self.driver.list_locations() if r.id == "nyc1"][0]
        floating_ip = self.driver.ex_create_floating_ip(nyc1)
        ret = self.driver.ex_delete_floating_ip(floating_ip)

        # The API returns 204 NO CONTENT if all is well.
        self.assertTrue(ret)

    def test_floating_ip_can_be_deleted_by_calling_delete_on_floating_ip_object(
        self,
    ):  # noqa: E501
        nyc1 = [r for r in self.driver.list_locations() if r.id == "nyc1"][0]
        floating_ip = self.driver.ex_create_floating_ip(nyc1)
        ret = floating_ip.delete()

        self.assertTrue(ret)

    def test_list_floating_ips(self):
        floating_ips = self.driver.ex_list_floating_ips()

        self.assertEqual(len(floating_ips), 2, "Wrong floating IPs count")

        floating_ip = floating_ips[0]
        self.assertEqual(floating_ip.id, "133.166.122.204")
        self.assertEqual(floating_ip.ip_address, "133.166.122.204")
        self.assertEqual(floating_ip.extra["region"]["slug"], "ams3")
        self.assertEqual(84155775, floating_ip.node_id)

    def test_get_floating_ip(self):
        floating_ip = self.driver.ex_get_floating_ip("133.166.122.204")

        self.assertEqual(floating_ip.id, "133.166.122.204")
        self.assertEqual(floating_ip.ip_address, "133.166.122.204")
        self.assertEqual(floating_ip.extra["region"]["slug"], "ams3")
        self.assertEqual(84155775, floating_ip.node_id)

    def test_ex_attach_floating_ip_to_node(self):
        node = self.driver.list_nodes()[0]
        floating_ip = self.driver.ex_get_floating_ip("133.166.122.204")

        ret = self.driver.ex_attach_floating_ip_to_node(node, floating_ip)

        self.assertTrue(ret)

    def test_ex_detach_floating_ip_from_node(self):
        node = self.driver.list_nodes()[0]
        floating_ip = self.driver.ex_get_floating_ip("154.138.103.175")

        ret = self.driver.ex_detach_floating_ip_from_node(node, floating_ip)

        self.assertTrue(ret)


class DigitalOceanMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("digitalocean_v2")

    def _v2_regions(self, method, url, body, headers):
        body = self.fixtures.load("list_locations.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_images(self, method, url, body, headers):
        body = self.fixtures.load("list_images.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_sizes(self, method, url, body, headers):
        body = self.fixtures.load("list_sizes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_droplets(self, method, url, body, headers):
        body = self.fixtures.load("list_nodes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_droplets_3164444(self, method, url, body, headers):
        body = self.fixtures.load("list_node.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_droplets_INVALID_IMAGE(self, method, url, body, headers):
        body = self.fixtures.load("error_invalid_image.json")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _v2_droplets_3164444_actions_REBOOT(self, method, url, body, headers):
        # reboot_node
        body = self.fixtures.load("reboot_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_DESTROY(self, method, url, body, headers):
        # destroy_node
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_droplets_3164444_actions_KERNELCHANGE(self, method, url, body, headers):
        # change_kernel
        body = self.fixtures.load("ex_change_kernel.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_actions_ENABLEIPV6(self, method, url, body, headers):
        # enable_ipv6
        body = self.fixtures.load("ex_enable_ipv6.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_actions_RENAME(self, method, url, body, headers):
        # rename_node
        body = self.fixtures.load("ex_rename_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_actions_SNAPSHOT(self, method, url, body, headers):
        # create_image
        body = self.fixtures.load("create_image.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_images_12345(self, method, url, body, headers):
        # get_image
        body = self.fixtures.load("get_image.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_images_12345_DESTROY(self, method, url, body, headers):
        # delete_image
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_droplets_3164444_actions_POWERON(self, method, url, body, headers):
        # ex_power_on_node
        body = self.fixtures.load("ex_power_on_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_actions_SHUTDOWN(self, method, url, body, headers):
        # ex_shutdown_node
        body = self.fixtures.load("ex_shutdown_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_3164444_actions_POWERCYCLE(self, method, url, body, headers):
        # ex_hard_reboot
        body = self.fixtures.load("ex_hard_reboot.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.OK])

    def _v2_droplets_3164444_actions_REBUILD(self, method, url, body, headers):
        # ex_rebuild_node
        body = self.fixtures.load("ex_rebuild_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.OK])

    def _v2_droplets_3164444_actions_RESIZE(self, method, url, body, headers):
        # ex_resize_node
        body = self.fixtures.load("ex_resize_node.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.OK])

    def _v2_account_keys(self, method, url, body, headers):
        body = self.fixtures.load("list_key_pairs.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_account_keys_7717(self, method, url, body, headers):
        # destroy_ssh_key
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_account_keys_CREATE(self, method, url, body, headers):
        # create_ssh_key
        body = self.fixtures.load("create_key_pair.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_droplets_UNAUTHORIZED(self, method, url, body, headers):
        body = self.fixtures.load("error.json")
        return (httplib.UNAUTHORIZED, body, {}, httplib.responses[httplib.UNAUTHORIZED])

    def _v2_droplets_PAGE_ONE(self, method, url, body, headers):
        body = self.fixtures.load("list_nodes_page_1.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_volumes(self, method, url, body, headers):
        body = self.fixtures.load("list_volumes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_volumes_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_volumes_empty.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_volumes_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("create_volume.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_volumes_actions_ATTACH(self, method, url, body, headers):
        body = self.fixtures.load("attach_volume.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _v2_volumes_DETACH(self, method, url, body, headers):
        body = self.fixtures.load("detach_volume.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_volumes_62766883_2c28_11e6_b8e6_000f53306ae1_DESTROY(self, method, url, body, headers):
        return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_volumes_62766883_2c28_11e6_b8e6_000f53306ae1_snapshots_CREATE(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_volume_snapshot.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _v2_volumes_62766883_2c28_11e6_b8e6_000f53306ae1_snapshots(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("list_volume_snapshots.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_snapshots_c0def940_9324_11e6_9a56_000f533176b1_DELETE(self, method, url, body, headers):
        return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_floating_ips(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("create_floating_ip.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "GET":
            body = self.fixtures.load("list_floating_ips.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:
            raise NotImplementedError()

    def _v2_floating_ips_167_138_123_111(self, method, url, body, headers):
        if method == "DELETE":
            body = ""
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])
        else:
            raise NotImplementedError()

    def _v2_floating_ips_133_166_122_204_actions(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("attach_floating_ip.json")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])
        else:
            raise NotImplementedError()

    def _v2_floating_ips_154_138_103_175_actions(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("detach_floating_ip.json")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])
        else:
            raise NotImplementedError()


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import LibcloudTestCase
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Object, Container
from libcloud.test.secrets import STORAGE_S3_PARAMS
from libcloud.storage.drivers.digitalocean_spaces import (
    DO_SPACES_HOSTS_BY_REGION,
    DOSpacesConnectionAWS2,
    DOSpacesConnectionAWS4,
    DigitalOceanSpacesStorageDriver,
)


class DigitalOceanSpacesTests(LibcloudTestCase):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = "nyc3.digitaloceanspaces.com"

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args, signature_version="2", host=self.default_host)

    def setUp(self):
        self.driver = self.create_driver()
        self.container = Container("test-container", {}, self.driver)
        self.object = Object("test-object", 1, "hash", {}, "meta_data", self.container, self.driver)

    def test_connection_class_type(self):
        res = self.driver.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, "driver.connectionCls does not match!")

    def test_connection_class_host(self):
        host = self.driver.connectionCls.host
        self.assertEqual(host, self.default_host)

    def test_container_enable_cdn_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.container.enable_cdn()

    def test_container_get_cdn_url_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.container.get_cdn_url()

    def test_object_enable_cdn_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.object.enable_cdn()

    def test_object_get_cdn_url_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.object.get_cdn_url()

    def test_invalid_signature_version(self):
        with self.assertRaises(ValueError):
            self.driver_type(*self.driver_args, signature_version="3", host=self.default_host)

    def test_invalid_region(self):
        with self.assertRaises(LibcloudError):
            self.driver_type(*self.driver_args, region="atlantis", host=self.default_host)


class DigitalOceanSpacesTests_v4(DigitalOceanSpacesTests):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = "nyc3.digitaloceanspaces.com"

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args, signature_version="4")

    def test_connection_class_type(self):
        res = self.driver.connectionCls is DOSpacesConnectionAWS4
        self.assertTrue(res, "driver.connectionCls does not match!")

    def test_connection_class_host(self):
        host = self.driver.connectionCls.host
        self.assertEqual(host, self.default_host)

    def test_valid_regions(self):
        for region, hostname in DO_SPACES_HOSTS_BY_REGION.items():
            driver = self.driver_type(*self.driver_args, region=region)
            self.assertEqual(driver.connectionCls.host, hostname)
            self.assertTrue(driver.connectionCls.host.startswith(region))


class DigitalOceanSpacesDoubleInstanceTests(LibcloudTestCase):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = "nyc3.digitaloceanspaces.com"
    alt_host = "ams3.digitaloceanspaces.com"

    def setUp(self):
        self.driver_v2 = self.driver_type(*self.driver_args, signature_version="2")
        self.driver_v4 = self.driver_type(*self.driver_args, signature_version="4", region="ams3")

    def test_connection_class_type(self):
        res = self.driver_v2.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, "driver.connectionCls does not match!")

        res = self.driver_v4.connectionCls is DOSpacesConnectionAWS4
        self.assertTrue(res, "driver.connectionCls does not match!")

        # Verify again that connection class hasn't been overridden when
        # instantiating a second driver class
        res = self.driver_v2.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, "driver.connectionCls does not match!")

    def test_connection_class_host(self):
        host = self.driver_v2.connectionCls.host
        self.assertEqual(host, self.default_host)

        host = self.driver_v4.connectionCls.host
        self.assertEqual(host, self.alt_host)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import random
import string
import hashlib

from libcloud.utils.py3 import httplib, urlencode, basestring
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import ProviderError, InvalidCredsError

SALT_CHARACTERS = string.ascii_letters + string.digits


class NFSNException(ProviderError):
    def __init__(self, value, http_code, code, driver=None):
        self.code = code
        super().__init__(value, http_code, driver)


class NFSNResponse(JsonResponse):
    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError("Invalid provider credentials")

        body = self.parse_body()

        if isinstance(body, basestring):
            return body + " (HTTP Code: %d)" % self.status

        error = body.get("error", None)
        debug = body.get("debug", None)
        # If we only have one of "error" or "debug", use the one that we have.
        # If we have both, use both, with a space character in between them.
        value = "No message specified"
        if error is not None:
            value = error
        if debug is not None:
            value = debug
        if error is not None and value is not None:
            value = error + " " + value
        value = value + " (HTTP Code: %d)" % self.status

        return value


class NFSNConnection(ConnectionUserAndKey):
    host = "api.nearlyfreespeech.net"
    responseCls = NFSNResponse
    allow_insecure = False

    def _header(self, action, data):
        """Build the contents of the X-NFSN-Authentication HTTP header. See
        https://members.nearlyfreespeech.net/wiki/API/Introduction for
        more explanation."""
        login = self.user_id
        timestamp = self._timestamp()
        salt = self._salt()
        api_key = self.key
        data = urlencode(data)
        data_hash = hashlib.sha1(data.encode("utf-8")).hexdigest()

        string = ";".join((login, timestamp, salt, api_key, action, data_hash))
        string_hash = hashlib.sha1(string.encode("utf-8")).hexdigest()

        return ";".join((login, timestamp, salt, string_hash))

    def request(self, action, params=None, data="", headers=None, method="GET"):
        """Add the X-NFSN-Authentication header to an HTTP request."""
        if not headers:
            headers = {}
        if not params:
            params = {}
        header = self._header(action, data)

        headers["X-NFSN-Authentication"] = header
        if method == "POST":
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        return ConnectionUserAndKey.request(self, action, params, data, headers, method)

    def encode_data(self, data):
        """NFSN expects the body to be regular key-value pairs that are not
        JSON-encoded."""
        if data:
            data = urlencode(data)
        return data

    def _salt(self):
        """Return a 16-character alphanumeric string."""
        r = random.SystemRandom()
        return "".join(r.choice(SALT_CHARACTERS) for _ in range(16))

    def _timestamp(self):
        """Return the current number of seconds since the Unix epoch,
        as a string."""
        return str(int(time.time()))



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

__all__ = [
    "SHOW_DEPRECATION_WARNING",
    "SHOW_IN_DEVELOPMENT_WARNING",
    "OLD_API_REMOVE_VERSION",
    "deprecated_warning",
    "in_development_warning",
]


SHOW_DEPRECATION_WARNING = True
SHOW_IN_DEVELOPMENT_WARNING = True
OLD_API_REMOVE_VERSION = "0.7.0"


def deprecated_warning(module):
    if SHOW_DEPRECATION_WARNING:
        warnings.warn(
            "This path has been deprecated and the module"
            ' is now available at "libcloud.compute.%s".'
            " This path will be fully removed in libcloud %s." % (module, OLD_API_REMOVE_VERSION),
            category=DeprecationWarning,
        )


def in_development_warning(module):
    if SHOW_IN_DEVELOPMENT_WARNING:
        warnings.warn(
            "The module %s is in development and your are advised "
            "against using it in production." % (module),
            category=FutureWarning,
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from email.utils import mktime_tz, parsedate_tz

__all__ = ["BaseHTTPError", "RateLimitReachedError", "exception_from_message"]


class BaseHTTPError(Exception):

    """
    The base exception class for all HTTP related exceptions.
    """

    def __init__(self, code, message, headers=None):
        self.code = code
        self.message = message
        self.headers = headers
        # preserve old exception behavior for tests that
        # look for e.args[0]
        super().__init__(message)

    def __str__(self):
        return self.message


class RateLimitReachedError(BaseHTTPError):
    """
    HTTP 429 - Rate limit: you've sent too many requests for this time period.
    """

    code = 429
    message = "%s Rate limit exceeded" % (code)

    def __init__(self, *args, **kwargs):
        headers = kwargs.pop("headers", None)
        super().__init__(self.code, self.message, headers)
        if self.headers is not None:
            self.retry_after = float(self.headers.get("retry-after", 0))
        else:
            self.retry_after = 0


_error_classes = [RateLimitReachedError]
_code_map = {c.code: c for c in _error_classes}


def exception_from_message(code, message, headers=None):
    """
    Return an instance of BaseHTTPException or subclass based on response code.

    If headers include Retry-After, RFC 2616 says that its value may be one of
    two formats: HTTP-date or delta-seconds, for example:

    Retry-After: Fri, 31 Dec 1999 23:59:59 GMT
    Retry-After: 120

    If Retry-After comes in HTTP-date, it'll be translated to a positive
    delta-seconds value when passing it to the exception constructor.

    Also, RFC 2616 says that Retry-After isn't just only applicable to 429
    HTTP status, but also to other responses, like 503 and 3xx.

    Usage::
        raise exception_from_message(code=self.status,
                                     message=self.parse_error(),
                                     headers=self.headers)
    """
    kwargs = {"code": code, "message": message, "headers": headers}

    if headers and "retry-after" in headers:
        http_date = parsedate_tz(headers["retry-after"])
        if http_date is not None:
            # Convert HTTP-date to delay-seconds
            delay = max(0, int(mktime_tz(http_date) - time.time()))
            headers["retry-after"] = str(delay)
    cls = _code_map.get(code, BaseHTTPError)
    return cls(**kwargs)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64

from libcloud.utils.py3 import b, httplib
from libcloud.common.base import JsonResponse, ConnectionUserAndKey


class PointDNSDNSResponse(JsonResponse):
    def success(self):
        """
        Determine if our request was successful.

        The meaning of this can be arbitrary; did we receive OK status? Did
        the node get created? Were we authenticated?

        :rtype: ``bool``
        :return: ``True`` or ``False``
        """
        # response.success() only checks for 200 and 201 codes. Should we
        # add 202?
        return self.status in [httplib.OK, httplib.CREATED, httplib.ACCEPTED]


class PointDNSConnection(ConnectionUserAndKey):
    host = "pointhq.com"
    responseCls = PointDNSDNSResponse

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request

        This method adds ``token`` to the request.
        """
        b64string = b("{}:{}".format(self.user_id, self.key))
        token = base64.b64encode(b64string)
        headers["Authorization"] = "Basic %s" % token
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"
        return headers



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["GoogleDNSDriver"]

# API docs: https://cloud.google.com/dns/api/v1
API_VERSION = "v1"

import re

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.common.google import GoogleResponse, GoogleBaseConnection, ResourceNotFoundError


class GoogleDNSResponse(GoogleResponse):
    pass


class GoogleDNSConnection(GoogleBaseConnection):
    host = "www.googleapis.com"
    responseCls = GoogleDNSResponse

    def __init__(
        self,
        user_id,
        key,
        secure,
        auth_type=None,
        credential_file=None,
        project=None,
        **kwargs,
    ):
        super().__init__(
            user_id,
            key,
            secure=secure,
            auth_type=auth_type,
            credential_file=credential_file,
            **kwargs,
        )
        self.request_path = "/dns/{}/projects/{}".format(API_VERSION, project)


class GoogleDNSDriver(DNSDriver):
    type = Provider.GOOGLE
    name = "Google DNS"
    connectionCls = GoogleDNSConnection
    website = "https://cloud.google.com/"

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.NS: "NS",
        RecordType.PTR: "PTR",
        RecordType.SOA: "SOA",
        RecordType.SPF: "SPF",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
        RecordType.CAA: "CAA",
    }

    def __init__(self, user_id, key, project=None, auth_type=None, scopes=None, **kwargs):
        self.auth_type = auth_type
        self.project = project
        self.scopes = scopes
        if not self.project:
            raise ValueError("Project name must be specified using " '"project" keyword.')
        super().__init__(user_id, key, **kwargs)

    def iterate_zones(self):
        """
        Return a generator to iterate over available zones.

        :rtype: ``generator`` of :class:`Zone`
        """
        return self._get_more("zones")

    def iterate_records(self, zone):
        """
        Return a generator to iterate over records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :rtype: ``generator`` of :class:`Record`
        """
        return self._get_more("records", zone=zone)

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        request = "/managedZones/%s" % (zone_id)

        try:
            response = self.connection.request(request, method="GET").object
        except ResourceNotFoundError:
            raise ZoneDoesNotExistError(value="", driver=self.connection.driver, zone_id=zone_id)

        return self._to_zone(response)

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        (record_type, record_name) = record_id.split(":", 1)

        params = {
            "name": record_name,
            "type": record_type,
        }

        request = "/managedZones/%s/rrsets" % (zone_id)

        try:
            response = self.connection.request(request, method="GET", params=params).object
        except ResourceNotFoundError:
            raise ZoneDoesNotExistError(value="", driver=self.connection.driver, zone_id=zone_id)

        if len(response["rrsets"]) > 0:
            zone = self.get_zone(zone_id)
            return self._to_record(response["rrsets"][0], zone)

        raise RecordDoesNotExistError(value="", driver=self.connection.driver, record_id=record_id)

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com.) with a \'.\'
                       at the end.
        :type domain: ``str``

        :param type: Zone type (master is the only one supported).
        :type  type: ``str``

        :param ttl: TTL for new records. (unused)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        name = None
        description = ""

        if extra:
            description = extra.get("description")
            name = extra.get("name")

        if name is None:
            name = self._cleanup_domain(domain)

        data = {
            "dnsName": domain,
            "name": name,
            "description": description,
        }

        request = "/managedZones"
        response = self.connection.request(request, method="POST", data=data).object
        return self._to_zone(response)

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        :param name: Record name fully qualified, with a \'.\' at the end.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes. (optional)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        ttl = data.get("ttl", 0)
        rrdatas = data.get("rrdatas", [])

        data = {"additions": [{"name": name, "type": type, "ttl": int(ttl), "rrdatas": rrdatas}]}
        request = "/managedZones/%s/changes" % (zone.id)
        response = self.connection.request(request, method="POST", data=data).object
        return self._to_record(response["additions"][0], zone)

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        request = "/managedZones/%s" % (zone.id)
        response = self.connection.request(request, method="DELETE")
        return response.success()

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        data = {
            "deletions": [
                {
                    "name": record.name,
                    "type": record.type,
                    "rrdatas": record.data["rrdatas"],
                    "ttl": record.data["ttl"],
                }
            ]
        }
        request = "/managedZones/%s/changes" % (record.zone.id)
        response = self.connection.request(request, method="POST", data=data)
        return response.success()

    def ex_bulk_record_changes(self, zone, records):
        """
        Bulk add and delete records.

        :param zone: Zone where the requested record changes are done.
        :type  zone: :class:`Zone`

        :param records: Dictionary of additions list or deletions list, or both
        of resourceRecordSets. For example:
            {'additions': [{'rrdatas': ['127.0.0.1'],
                            'kind': 'dns#resourceRecordSet',
                            'type': 'A',
                            'name': 'www.example.com.',
                            'ttl': '300'}],
             'deletions': [{'rrdatas': ['127.0.0.1'],
                            'kind': 'dns#resourceRecordSet',
                            'type': 'A',
                            'name': 'www2.example.com.',
                            'ttl': '300'}]}
        :type  records: ``dict``

        :return: A dictionary of Record additions and deletions.
        :rtype: ``dict`` of additions and deletions of :class:`Record`
        """

        request = "/managedZones/%s/changes" % (zone.id)
        response = self.connection.request(request, method="POST", data=records).object
        response = response or {}

        response_data = {
            "additions": self._to_records(response.get("additions", []), zone),
            "deletions": self._to_records(response.get("deletions", []), zone),
        }

        return response_data

    def _get_more(self, rtype, **kwargs):
        last_key = None
        exhausted = False
        while not exhausted:
            items, last_key, exhausted = self._get_data(rtype, last_key, **kwargs)
            yield from items

    def _get_data(self, rtype, last_key, **kwargs):
        params = {}

        if last_key:
            params["pageToken"] = last_key

        if rtype == "zones":
            request = "/managedZones"
            transform_func = self._to_zones
            r_key = "managedZones"
        elif rtype == "records":
            zone = kwargs["zone"]
            request = "/managedZones/%s/rrsets" % (zone.id)
            transform_func = self._to_records
            r_key = "rrsets"

        response = self.connection.request(
            request,
            method="GET",
            params=params,
        )

        if response.success():
            nextpage = response.object.get("nextPageToken", None)
            items = transform_func(response.object.get(r_key), **kwargs)
            exhausted = False if nextpage is not None else True
            return items, nextpage, exhausted
        else:
            return [], None, True

    def _ex_connection_class_kwargs(self):
        return {
            "auth_type": self.auth_type,
            "project": self.project,
            "scopes": self.scopes,
        }

    def _to_zones(self, response):
        zones = []
        for r in response:
            zones.append(self._to_zone(r))
        return zones

    def _to_zone(self, r):
        extra = {}

        if "description" in r:
            extra["description"] = r.get("description")

        extra["creationTime"] = r.get("creationTime")
        extra["nameServers"] = r.get("nameServers")
        extra["id"] = r.get("id")

        return Zone(
            id=r["name"],
            domain=r["dnsName"],
            type="master",
            ttl=0,
            driver=self,
            extra=extra,
        )

    def _to_records(self, response, zone):
        records = []
        for r in response:
            records.append(self._to_record(r, zone))
        return records

    def _to_record(self, r, zone):
        record_id = "{}:{}".format(r["type"], r["name"])
        return Record(
            id=record_id,
            name=r["name"],
            type=r["type"],
            data=r,
            zone=zone,
            driver=self,
            ttl=r.get("ttl", None),
            extra={},
        )

    def _cleanup_domain(self, domain):
        # name can only contain lower case alphanumeric characters and hyphens
        domain = re.sub(r"[^a-zA-Z0-9-]", "-", domain)
        if domain[-1] == "-":
            domain = domain[:-1]
        return domain



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from shlex import quote as pquote
from xml.dom.minidom import parseString

from libcloud.utils.py3 import _real_unicode as u
from libcloud.utils.py3 import ensure_string
from libcloud.utils.misc import lowercase_keys
from libcloud.common.base import LibcloudConnection, HttpLibResponseProxy

try:
    import simplejson as json  # type: ignore
except Exception:
    import json  # type: ignore


class LoggingConnection(LibcloudConnection):
    """
    Debug class to log all HTTP(s) requests as they could be made
    with the curl command.

    :cvar log: file-like object that logs entries are written to.
    """

    protocol = "https"

    log = None
    http_proxy_used = False

    def _log_response(self, r):
        rv = "# -------- begin %d:%d response ----------\n" % (id(self), id(r))
        ht = ""
        v = r.version
        if r.version == 10:
            v = "HTTP/1.0"
        if r.version == 11:
            v = "HTTP/1.1"
        ht += "{} {} {}\r\n".format(v, r.status, r.reason)
        body = r.read()
        for h in r.getheaders():
            ht += "{}: {}\r\n".format(h[0].title(), h[1])
        ht += "\r\n"

        headers = lowercase_keys(dict(r.getheaders()))

        content_type = headers.get("content-type", None)

        pretty_print = os.environ.get("LIBCLOUD_DEBUG_PRETTY_PRINT_RESPONSE", False)

        if pretty_print and content_type == "application/json":
            try:
                body = json.loads(ensure_string(body))
                body = json.dumps(body, sort_keys=True, indent=4)
            except Exception:
                # Invalid JSON or server is lying about content-type
                pass
        elif pretty_print and content_type in ["text/xml", "application/xml"]:
            try:
                elem = parseString(body.decode("utf-8"))
                body = elem.toprettyxml()
            except Exception:
                # Invalid XML
                pass

        ht += ensure_string(body)

        rv += ht
        rv += "\n# -------- end %d:%d response ----------\n" % (id(self), id(r))

        return rv

    def _log_curl(self, method, url, body, headers):
        cmd = ["curl"]

        if self.http_proxy_used:
            if self.proxy_username and self.proxy_password:
                proxy_url = "{}://{}:{}@{}:{}".format(
                    self.proxy_scheme,
                    self.proxy_username,
                    self.proxy_password,
                    self.proxy_host,
                    self.proxy_port,
                )
            else:
                proxy_url = "{}://{}:{}".format(
                    self.proxy_scheme,
                    self.proxy_host,
                    self.proxy_port,
                )
            proxy_url = pquote(proxy_url)
            cmd.extend(["--proxy", proxy_url])

        cmd.extend(["-i"])

        if method.lower() == "head":
            # HEAD method need special handling
            cmd.extend(["--head"])
        else:
            cmd.extend(["-X", pquote(method)])

        for h in headers:
            cmd.extend(["-H", pquote("{}: {}".format(h, headers[h]))])

        cert_file = getattr(self, "cert_file", None)

        if cert_file:
            cmd.extend(["--cert", pquote(cert_file)])

        # TODO: in python 2.6, body can be a file-like object.
        if body is not None and len(body) > 0:
            if isinstance(body, (bytearray, bytes)):
                body = body.decode("utf-8")

            cmd.extend(["--data-binary", pquote(body)])

        cmd.extend(["--compress"])
        cmd.extend([pquote("{}{}".format(self.host, url))])
        return " ".join(cmd)

    def getresponse(self):
        original_response = LibcloudConnection.getresponse(self)
        if self.log is not None:
            rv = self._log_response(HttpLibResponseProxy(original_response))
            self.log.write(u(rv + "\n"))
            self.log.flush()
        return original_response

    def request(self, method, url, body=None, headers=None, **kwargs):
        headers.update({"X-LC-Request-ID": str(id(self))})
        if self.log is not None:
            pre = "# -------- begin %d request ----------\n" % id(self)
            self.log.write(u(pre + self._log_curl(method, url, body, headers) + "\n"))
            self.log.flush()
        return LibcloudConnection.request(self, method, url, body, headers)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Digital Ocean DNS Driver
"""

__all__ = ["DigitalOceanDNSDriver"]

import json

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType
from libcloud.utils.py3 import httplib
from libcloud.common.digitalocean import DigitalOcean_v2_BaseDriver, DigitalOcean_v2_Connection


class DigitalOceanDNSDriver(DigitalOcean_v2_BaseDriver, DNSDriver):
    connectionCls = DigitalOcean_v2_Connection
    type = Provider.DIGITAL_OCEAN
    name = "DigitalOcean"
    website = "https://www.digitalocean.com"

    RECORD_TYPE_MAP = {
        RecordType.NS: "NS",
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.TXT: "TXT",
        RecordType.SRV: "SRV",
    }

    def list_zones(self):
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        data = self._paginated_request("/v2/domains", "domains")
        return list(map(self._to_zone, data))

    def list_records(self, zone):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        data = self._paginated_request("/v2/domains/%s/records" % (zone.id), "domain_records")
        # TODO: Not use list comprehension to add zone to record for proper data map
        #       functionality? This passes a reference to zone for each data currently
        #       to _to_record which returns a Record. map() does not take keywords
        return list(map(self._to_record, data, [zone for z in data]))

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        data = self.connection.request("/v2/domains/%s" % (zone_id)).object["domain"]

        return self._to_zone(data)

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        data = self.connection.request(
            "/v2/domains/{}/records/{}".format(zone_id, record_id)
        ).object["domain_record"]

        # TODO: Any way of not using get_zone which polls the API again
        #       without breaking the DNSDriver.get_record parameters?
        return self._to_record(data, self.get_zone(zone_id))

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type (master / slave) (does nothing).
        :type  type: ``str``

        :param ttl: TTL for new records. (does nothing)
        :type  ttl: ``int``

        :param extra: Extra attributes (to set ip). (optional)
                      Note: This can be used to set the default A record with
                      {"ip" : "IP.AD.DR.ESS"} otherwise 127.0.0.1 is used
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        params = {"name": domain}
        try:
            params["ip_address"] = extra["ip"]
        except Exception:
            params["ip_address"] = "127.0.0.1"

        res = self.connection.request("/v2/domains", data=json.dumps(params), method="POST")

        return Zone(
            id=res.object["domain"]["name"],
            domain=res.object["domain"]["name"],
            type="master",
            ttl=1800,
            driver=self,
            extra={},
        )

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes for MX and SRV. (Depends on record)
                      {"priority" : 0, "port" : 443, "weight" : 100}
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        params = {"type": self.RECORD_TYPE_MAP[type], "name": name, "data": data}
        if extra:
            try:
                params["priority"] = extra["priority"]
            except KeyError:
                params["priority"] = None
            try:
                params["port"] = extra["port"]
            except KeyError:
                params["port"] = None
            try:
                params["weight"] = extra["weight"]
            except KeyError:
                params["weight"] = None

            if "ttl" in extra:
                params["ttl"] = extra["ttl"]

        res = self.connection.request(
            "/v2/domains/%s/records" % zone.id, data=json.dumps(params), method="POST"
        )

        return Record(
            id=res.object["domain_record"]["id"],
            name=res.object["domain_record"]["name"],
            type=type,
            data=data,
            zone=zone,
            ttl=res.object["domain_record"].get("ttl", None),
            driver=self,
            extra=extra,
        )

    def update_record(self, record, name=None, type=None, data=None, extra=None):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name without the domain name (e.g. www). (Ignored)
                     Note: The value is pulled from the record being updated
        :type  name: ``str``

        :param type: DNS record type (A, AAAA, ...). (Ignored)
                     Note: Updating records does not support changing type
                     so this value is ignored
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        params = {"type": record.type, "name": record.name, "data": data}
        if data is None:
            params["data"] = record.data
        if extra:
            try:
                params["priority"] = extra["priority"]
            except KeyError:
                params["priority"] = None
            try:
                params["port"] = extra["port"]
            except KeyError:
                params["port"] = None
            try:
                params["weight"] = extra["weight"]
            except KeyError:
                params["weight"] = None

            if "ttl" in extra:
                params["ttl"] = extra["ttl"]

        res = self.connection.request(
            "/v2/domains/{}/records/{}".format(record.zone.id, record.id),
            data=json.dumps(params),
            method="PUT",
        )

        return Record(
            id=res.object["domain_record"]["id"],
            name=res.object["domain_record"]["name"],
            type=record.type,
            data=data,
            zone=record.zone,
            ttl=res.object["domain_record"].get("ttl", None),
            driver=self,
            extra=extra,
        )

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        params = {}

        res = self.connection.request("/v2/domains/%s" % zone.id, params=params, method="DELETE")

        return res.status == httplib.NO_CONTENT

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        params = {}

        res = self.connection.request(
            "/v2/domains/{}/records/{}".format(record.zone.id, record.id),
            params=params,
            method="DELETE",
        )
        return res.status == httplib.NO_CONTENT

    def _to_record(self, data, zone=None):
        extra = {
            "port": data["port"],
            "priority": data["priority"],
            "weight": data["weight"],
        }
        return Record(
            id=data["id"],
            name=data["name"],
            type=self._string_to_record_type(data["type"]),
            data=data["data"],
            zone=zone,
            ttl=data.get("ttl", None),
            driver=self,
            extra=extra,
        )

    def _to_zone(self, data):
        extra = {"zone_file": data["zone_file"]}
        return Zone(
            id=data["name"],
            domain=data["name"],
            type="master",
            ttl=data["ttl"],
            driver=self,
            extra=extra,
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
from unittest import TestCase

from libcloud.compute.types import (
    Type,
    Provider,
    NodeState,
    StorageVolumeState,
    VolumeSnapshotState,
)


class TestType(Type):
    INUSE = "inuse"
    NOTINUSE = "NOTINUSE"


class TestTestType(TestCase):
    model = TestType

    def test_provider_tostring(self):
        self.assertEqual(Provider.tostring(TestType.INUSE), "INUSE")
        self.assertEqual(Provider.tostring(TestType.NOTINUSE), "NOTINUSE")

    def test_provider_fromstring(self):
        self.assertEqual(TestType.fromstring("inuse"), TestType.INUSE)
        self.assertEqual(TestType.fromstring("NOTINUSE"), TestType.NOTINUSE)

    def test_provider_fromstring_caseinsensitive(self):
        self.assertEqual(TestType.fromstring("INUSE"), TestType.INUSE)
        self.assertEqual(TestType.fromstring("notinuse"), TestType.NOTINUSE)

    def test_compare_as_string(self):
        self.assertTrue(TestType.INUSE == "inuse")
        self.assertFalse(TestType.INUSE == "bar")


class TestProvider(TestCase):
    def test_provider_tostring(self):
        self.assertEqual(Provider.tostring(Provider.RACKSPACE), "RACKSPACE")

    def test_provider_fromstring(self):
        self.assertEqual(Provider.fromstring("rackspace"), Provider.RACKSPACE)


class TestNodeState(TestCase):
    def test_nodestate_tostring(self):
        self.assertEqual(NodeState.tostring(NodeState.RUNNING), "RUNNING")

    def test_nodestate_fromstring(self):
        self.assertEqual(NodeState.fromstring("running"), NodeState.RUNNING)


class TestStorageVolumeState(TestCase):
    def test_storagevolumestate_tostring(self):
        self.assertEqual(StorageVolumeState.tostring(StorageVolumeState.AVAILABLE), "AVAILABLE")

    def test_storagevolumestate_fromstring(self):
        self.assertEqual(StorageVolumeState.fromstring("available"), StorageVolumeState.AVAILABLE)


class TestVolumeSnapshotState(TestCase):
    def test_volumesnapshotstate_tostring(self):
        self.assertEqual(VolumeSnapshotState.tostring(VolumeSnapshotState.AVAILABLE), "AVAILABLE")

    def test_volumesnapshotstate_fromstring(self):
        self.assertEqual(VolumeSnapshotState.fromstring("available"), VolumeSnapshotState.AVAILABLE)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.compute.base import Node
from libcloud.test.secrets import LB_SLB_PARAMS
from libcloud.compute.types import NodeState
from libcloud.loadbalancer.base import Member, Algorithm
from libcloud.loadbalancer.types import State
from libcloud.test.file_fixtures import LoadBalancerFileFixtures
from libcloud.loadbalancer.drivers.slb import (
    SLBDriver,
    SLBLoadBalancerTcpListener,
    SLBLoadBalancerUdpListener,
    SLBLoadBalancerHttpListener,
    SLBLoadBalancerHttpsListener,
)


class SLBDriverTestCases(unittest.TestCase):
    region = LB_SLB_PARAMS[2]

    def setUp(self):
        SLBMockHttp.test = self
        SLBDriver.connectionCls.conn_class = SLBMockHttp
        SLBMockHttp.type = None
        SLBMockHttp.use_param = "Action"

        self.driver = SLBDriver(*LB_SLB_PARAMS)

    def test_list_protocols(self):
        protocols = self.driver.list_protocols()
        self.assertEqual(4, len(protocols))
        expected = ["tcp", "udp", "http", "https"]
        diff = set(expected) - set(protocols)
        self.assertEqual(0, len(diff))

    def test_list_balancers(self):
        balancers = self.driver.list_balancers()

        self.assertEqual(len(balancers), 1)
        balancer = balancers[0]
        self.assertEqual("15229f88562-cn-hangzhou-dg-a01", balancer.id)
        self.assertEqual("abc", balancer.name)
        self.assertEqual(State.RUNNING, balancer.state)
        self.assertEqual("120.27.186.149", balancer.ip)
        self.assertTrue(balancer.port is None)
        self.assertEqual(self.driver, balancer.driver)
        expected_extra = {
            "create_timestamp": 1452403099000,
            "address_type": "internet",
            "region_id": "cn-hangzhou-dg-a01",
            "region_id_alias": "cn-hangzhou",
            "create_time": "2016-01-10T13:18Z",
            "master_zone_id": "cn-hangzhou-d",
            "slave_zone_id": "cn-hangzhou-b",
            "network_type": "classic",
        }
        self._validate_extras(expected_extra, balancer.extra)

    def _validate_extras(self, expected, actual):
        self.assertTrue(actual is not None)
        for key, value in iter(expected.items()):
            self.assertTrue(key in actual)
            self.assertEqual(
                value,
                actual[key],
                (
                    "extra %(key)s not equal, "
                    'expected: "%(expected)s", '
                    'actual: "%(actual)s"' % {"key": key, "expected": value, "actual": actual[key]}
                ),
            )

    def test_list_balancers_with_ids(self):
        SLBMockHttp.type = "list_balancers_ids"
        self.balancer_ids = ["id1", "id2"]
        balancers = self.driver.list_balancers(ex_balancer_ids=self.balancer_ids)
        self.assertTrue(balancers is not None)

    def test_list_balancers_with_ex_filters(self):
        SLBMockHttp.type = "list_balancers_filters"
        self.ex_filters = {"AddressType": "internet"}
        balancers = self.driver.list_balancers(ex_filters=self.ex_filters)
        self.assertTrue(balancers is not None)

    def test_get_balancer(self):
        SLBMockHttp.type = "get_balancer"
        balancer = self.driver.get_balancer(balancer_id="tests")

        self.assertEqual(balancer.id, "15229f88562-cn-hangzhou-dg-a01")
        self.assertEqual(balancer.name, "abc")
        self.assertEqual(balancer.state, State.RUNNING)

    def test_destroy_balancer(self):
        balancer = self.driver.get_balancer(balancer_id="tests")

        self.assertTrue(self.driver.destroy_balancer(balancer))

    def test_create_balancer(self):
        self.name = "balancer1"
        self.port = 80
        self.protocol = "http"
        self.algorithm = Algorithm.WEIGHTED_ROUND_ROBIN
        self.extra = {
            "ex_address_type": "internet",
            "ex_internet_charge_type": "paybytraffic",
            "ex_bandwidth": 1,
            "ex_master_zone_id": "cn-hangzhou-d",
            "ex_slave_zone_id": "cn-hangzhou-b",
            "StickySession": "on",
            "HealthCheck": "on",
        }
        self.members = [Member("node1", None, None)]

        balancer = self.driver.create_balancer(
            name=self.name,
            port=self.port,
            protocol=self.protocol,
            algorithm=self.algorithm,
            members=self.members,
            **self.extra,
        )

        self.assertEqual(balancer.name, self.name)
        self.assertEqual(balancer.port, self.port)
        self.assertEqual(balancer.state, State.UNKNOWN)

    def test_create_balancer_no_port_exception(self):
        self.assertRaises(
            AttributeError,
            self.driver.create_balancer,
            None,
            None,
            "http",
            Algorithm.WEIGHTED_ROUND_ROBIN,
            None,
        )

    def test_create_balancer_unsupport_protocol_exception(self):
        self.assertRaises(
            AttributeError,
            self.driver.create_balancer,
            None,
            443,
            "ssl",
            Algorithm.WEIGHTED_ROUND_ROBIN,
            None,
        )

    def test_create_balancer_multiple_member_ports_exception(self):
        members = [Member("m1", "1.2.3.4", 80), Member("m2", "1.2.3.5", 81)]
        self.assertRaises(
            AttributeError,
            self.driver.create_balancer,
            None,
            80,
            "http",
            Algorithm.WEIGHTED_ROUND_ROBIN,
            members,
        )

    def test_create_balancer_bandwidth_value_error(self):
        self.assertRaises(
            AttributeError,
            self.driver.create_balancer,
            None,
            80,
            "http",
            Algorithm.WEIGHTED_ROUND_ROBIN,
            None,
            ex_bandwidth="NAN",
        )

    def test_create_balancer_paybybandwidth_without_bandwidth_exception(self):
        self.assertRaises(
            AttributeError,
            self.driver.create_balancer,
            None,
            80,
            "http",
            Algorithm.WEIGHTED_ROUND_ROBIN,
            None,
            ex_internet_charge_type="paybybandwidth",
        )

    def test_balancer_list_members(self):
        balancer = self.driver.get_balancer(balancer_id="tests")
        members = balancer.list_members()

        self.assertEqual(len(members), 1)
        self.assertEqual(members[0].balancer, balancer)
        self.assertEqual("i-23tshnsdq", members[0].id)

    def test_balancer_list_listeners(self):
        balancer = self.driver.get_balancer(balancer_id="tests")
        listeners = self.driver.ex_list_listeners(balancer)

        self.assertEqual(1, len(listeners))
        listener = listeners[0]
        self.assertEqual("80", listener.port)

    def test_balancer_detach_member(self):
        self.balancer = self.driver.get_balancer(balancer_id="tests")
        self.member = Member("i-23tshnsdq", None, None)

        self.assertTrue(self.balancer.detach_member(self.member))

    def test_balancer_attach_compute_node(self):
        SLBMockHttp.type = "attach_compute_node"
        self.balancer = self.driver.get_balancer(balancer_id="tests")
        self.node = Node(
            id="node1",
            name="node-name",
            state=NodeState.RUNNING,
            public_ips=["1.2.3.4"],
            private_ips=["4.3.2.1"],
            driver=self.driver,
        )
        member = self.driver.balancer_attach_compute_node(self.balancer, self.node)
        self.assertEqual(self.node.id, member.id)
        self.assertEqual(self.node.public_ips[0], member.ip)
        self.assertEqual(self.balancer.port, member.port)

    def test_ex_create_listener(self):
        SLBMockHttp.type = "create_listener"
        self.balancer = self.driver.get_balancer(balancer_id="tests")
        self.backend_port = 80
        self.protocol = "http"
        self.algorithm = Algorithm.WEIGHTED_ROUND_ROBIN
        self.bandwidth = 1
        self.extra = {"StickySession": "off", "HealthCheck": "off"}
        self.assertTrue(
            self.driver.ex_create_listener(
                self.balancer,
                self.backend_port,
                self.protocol,
                self.algorithm,
                self.bandwidth,
                **self.extra,
            )
        )

    def test_ex_create_listener_override_port(self):
        SLBMockHttp.type = "create_listener_override_port"
        self.balancer = self.driver.get_balancer(balancer_id="tests")
        self.backend_port = 80
        self.protocol = "http"
        self.algorithm = Algorithm.WEIGHTED_ROUND_ROBIN
        self.bandwidth = 1
        self.extra = {
            "StickySession": "off",
            "HealthCheck": "off",
            "ListenerPort": 8080,
        }
        self.assertTrue(
            self.driver.ex_create_listener(
                self.balancer,
                self.backend_port,
                self.protocol,
                self.algorithm,
                self.bandwidth,
                **self.extra,
            )
        )

    def test_ex_start_listener(self):
        SLBMockHttp.type = "start_listener"
        balancer = self.driver.get_balancer(balancer_id="tests")
        self.port = 80
        self.assertTrue(self.driver.ex_start_listener(balancer, self.port))

    def test_ex_stop_listener(self):
        SLBMockHttp.type = "stop_listener"
        balancer = self.driver.get_balancer(balancer_id="tests")
        self.port = 80
        self.assertTrue(self.driver.ex_stop_listener(balancer, self.port))

    def test_ex_upload_certificate(self):
        self.name = "cert1"
        self.cert = "cert-data"
        self.key = "key-data"
        certificate = self.driver.ex_upload_certificate(self.name, self.cert, self.key)
        self.assertEqual(self.name, certificate.name)
        self.assertEqual("01:DF:AB:CD", certificate.fingerprint)

    def test_ex_list_certificates(self):
        certs = self.driver.ex_list_certificates()
        self.assertEqual(2, len(certs))
        cert = certs[0]
        self.assertEqual("139a00604ad-cn-east-hangzhou-01", cert.id)
        self.assertEqual("abe", cert.name)
        self.assertEqual("A:B:E", cert.fingerprint)

    def test_ex_list_certificates_ids(self):
        SLBMockHttp.type = "list_certificates_ids"
        self.cert_ids = ["cert1", "cert2"]
        certs = self.driver.ex_list_certificates(certificate_ids=self.cert_ids)
        self.assertTrue(certs is not None)

    def test_ex_delete_certificate(self):
        self.cert_id = "cert1"
        self.assertTrue(self.driver.ex_delete_certificate(self.cert_id))

    def test_ex_set_certificate_name(self):
        self.cert_id = "cert1"
        self.cert_name = "cert-name"
        self.assertTrue(self.driver.ex_set_certificate_name(self.cert_id, self.cert_name))


class SLBMockHttp(MockHttp, unittest.TestCase):
    fixtures = LoadBalancerFileFixtures("slb")

    def _DescribeLoadBalancers(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _list_balancers_ids_DescribeLoadBalancers(self, method, url, body, headers):
        params = {"LoadBalancerId": ",".join(self.test.balancer_ids)}
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _list_balancers_filters_DescribeLoadBalancers(self, method, url, body, headers):
        params = {"AddressType": "internet"}
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _get_balancer_DescribeLoadBalancers(self, method, url, body, headers):
        params = {"LoadBalancerId": "tests"}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeLoadBalancers(method, url, body, headers)

    def _DeleteLoadBalancer(self, method, url, body, headers):
        params = {"LoadBalancerId": "15229f88562-cn-hangzhou-dg-a01"}
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("delete_load_balancer.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _DescribeLoadBalancerAttribute(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancer_attribute.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _CreateLoadBalancer(self, method, url, body, headers):
        params = {"RegionId": self.test.region, "LoadBalancerName": self.test.name}
        balancer_keys = {
            "AddressType": "ex_address_type",
            "InternetChargeType": "ex_internet_charge_type",
            "Bandwidth": "ex_bandwidth",
            "MasterZoneId": "ex_master_zone_id",
            "SlaveZoneId": "ex_slave_zone_id",
        }
        for key in balancer_keys:
            params[key] = str(self.test.extra[balancer_keys[key]])

        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("create_load_balancer.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _AddBackendServers(self, method, url, body, headers):
        _id = self.test.members[0].id
        self.assertTrue("ServerId" in url and _id in url)
        self.assertTrue("Weight" in url and "100" in url)
        body = self.fixtures.load("add_backend_servers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _CreateLoadBalancerHTTPListener(self, method, url, body, headers):
        body = self.fixtures.load("create_load_balancer_http_listener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _StartLoadBalancerListener(self, method, url, body, headers):
        body = self.fixtures.load("start_load_balancer_listener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _RemoveBackendServers(self, method, url, body, headers):
        _id = self.test.member.id
        servers_json = '["%s"]' % _id
        params = {
            "LoadBalancerId": self.test.balancer.id,
            "BackendServers": servers_json,
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("add_backend_servers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _attach_compute_node_DescribeLoadBalancers(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _attach_compute_node_AddBackendServers(self, method, url, body, headers):
        _id = self.test.node.id
        self.assertTrue("ServerId" in url and _id in url)
        self.assertTrue("Weight" in url and "100" in url)
        body = self.fixtures.load("add_backend_servers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _create_listener_CreateLoadBalancerHTTPListener(self, method, url, body, headers):
        params = {
            "LoadBalancerId": self.test.balancer.id,
            "RegionId": self.test.region,
            "ListenerPort": str(self.test.balancer.port),
            "BackendServerPort": str(self.test.backend_port),
            "Scheduler": "wrr",
            "Bandwidth": "1",
            "StickySession": "off",
            "HealthCheck": "off",
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("create_load_balancer_http_listener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _create_listener_DescribeLoadBalancers(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _create_listener_override_port_CreateLoadBalancerHTTPListener(
        self, method, url, body, headers
    ):
        params = {
            "LoadBalancerId": self.test.balancer.id,
            "RegionId": self.test.region,
            "ListenerPort": str(self.test.extra["ListenerPort"]),
            "BackendServerPort": str(self.test.backend_port),
            "Scheduler": "wrr",
            "Bandwidth": "1",
            "StickySession": "off",
            "HealthCheck": "off",
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("create_load_balancer_http_listener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _create_listener_override_port_DescribeLoadBalancers(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _start_listener_DescribeLoadBalancers(self, method, url, body, headers):
        return self._DescribeLoadBalancers(method, url, body, headers)

    def _start_listener_StartLoadBalancerListener(self, method, url, body, headers):
        params = {"ListenerPort": str(self.test.port)}
        self.assertUrlContainsQueryParams(url, params)
        return self._StartLoadBalancerListener(method, url, body, headers)

    def _stop_listener_DescribeLoadBalancers(self, method, url, body, headers):
        return self._DescribeLoadBalancers(method, url, body, headers)

    def _stop_listener_StopLoadBalancerListener(self, method, url, body, headers):
        params = {"ListenerPort": str(self.test.port)}
        self.assertUrlContainsQueryParams(url, params)
        return self._StartLoadBalancerListener(method, url, body, headers)

    def _UploadServerCertificate(self, method, url, body, headers):
        body = self.fixtures.load("upload_server_certificate.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _DescribeServerCertificates(self, method, url, body, headers):
        body = self.fixtures.load("describe_server_certificates.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _list_certificates_ids_DescribeServerCertificates(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ServerCertificateId": ",".join(self.test.cert_ids),
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("describe_server_certificates.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _DeleteServerCertificate(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ServerCertificateId": self.test.cert_id,
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("delete_server_certificate.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _SetServerCertificateName(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ServerCertificateId": self.test.cert_id,
            "ServerCertificateName": self.test.cert_name,
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("set_server_certificate_name.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


class AssertDictMixin:
    def assert_dict_equals(self, expected, actual):
        expected_keys = set(expected.keys())
        actual_keys = set(actual.keys())
        self.assertEqual(len(expected_keys), len(actual_keys))
        self.assertEqual(0, len(expected_keys - actual_keys))
        for key in expected:
            self.assertEqual(expected[key], actual[key])


class SLBLoadBalancerHttpListenerTestCase(unittest.TestCase, AssertDictMixin):
    def setUp(self):
        self.listener = SLBLoadBalancerHttpListener.create(
            80,
            8080,
            Algorithm.WEIGHTED_ROUND_ROBIN,
            1,
            extra={
                "StickySession": "on",
                "StickySessionType": "insert",
                "HealthCheck": "on",
            },
        )

    def test_get_required_params(self):
        expected = {
            "Action": "CreateLoadBalancerHTTPListener",
            "ListenerPort": 80,
            "BackendServerPort": 8080,
            "Scheduler": "wrr",
            "Bandwidth": 1,
            "StickySession": "on",
            "HealthCheck": "on",
        }
        self.assert_dict_equals(expected, self.listener.get_required_params())

    def test_get_optional_params(self):
        expected = {"StickySessionType": "insert"}
        self.assert_dict_equals(expected, self.listener.get_optional_params())

    def test_repr(self):
        self.assertTrue("SLBLoadBalancerHttpListener" in str(self.listener))


class SLBLoadBalancerHttpsListenerTestCase(unittest.TestCase, AssertDictMixin):
    def setUp(self):
        self.listener = SLBLoadBalancerHttpsListener.create(
            80,
            8080,
            Algorithm.WEIGHTED_ROUND_ROBIN,
            1,
            extra={
                "StickySession": "on",
                "StickySessionType": "insert",
                "HealthCheck": "on",
                "ServerCertificateId": "fake-cert1",
            },
        )

    def test_get_required_params(self):
        expected = {
            "Action": "CreateLoadBalancerHTTPSListener",
            "ListenerPort": 80,
            "BackendServerPort": 8080,
            "Scheduler": "wrr",
            "Bandwidth": 1,
            "StickySession": "on",
            "HealthCheck": "on",
            "ServerCertificateId": "fake-cert1",
        }
        self.assert_dict_equals(expected, self.listener.get_required_params())

    def test_get_optional_params(self):
        expected = {"StickySessionType": "insert"}
        self.assert_dict_equals(expected, self.listener.get_optional_params())


class SLBLoadBalancerTcpListenerTestCase(unittest.TestCase, AssertDictMixin):
    def setUp(self):
        self.listener = SLBLoadBalancerTcpListener.create(
            80,
            8080,
            Algorithm.WEIGHTED_ROUND_ROBIN,
            1,
            extra={"PersistenceTimeout": 0, "HealthCheckDomain": ""},
        )

    def test_get_required_params(self):
        expected = {
            "Action": "CreateLoadBalancerTCPListener",
            "ListenerPort": 80,
            "BackendServerPort": 8080,
            "Scheduler": "wrr",
            "Bandwidth": 1,
        }
        self.assert_dict_equals(expected, self.listener.get_required_params())

    def test_get_optional_params(self):
        expected = {"PersistenceTimeout": 0, "HealthCheckDomain": ""}
        self.assert_dict_equals(expected, self.listener.get_optional_params())


class SLBLoadBalancerUdpListenerTestCase(unittest.TestCase, AssertDictMixin):
    def setUp(self):
        self.listener = SLBLoadBalancerUdpListener.create(
            80, 8080, Algorithm.WEIGHTED_ROUND_ROBIN, 1, extra={"PersistenceTimeout": 0}
        )

    def test_get_required_params(self):
        expected = {
            "Action": "CreateLoadBalancerUDPListener",
            "ListenerPort": 80,
            "BackendServerPort": 8080,
            "Scheduler": "wrr",
            "Bandwidth": 1,
        }
        self.assert_dict_equals(expected, self.listener.get_required_params())

    def test_get_optional_params(self):
        expected = {"PersistenceTimeout": 0}
        self.assert_dict_equals(expected, self.listener.get_optional_params())


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.types import LibcloudError

__all__ = [
    "Provider",
    "ContainerError",
    "ObjectError",
    "ContainerAlreadyExistsError",
    "ContainerDoesNotExistError",
    "ContainerIsNotEmptyError",
    "ObjectDoesNotExistError",
    "ObjectHashMismatchError",
    "InvalidContainerNameError",
    "OLD_CONSTANT_TO_NEW_MAPPING",
]


class Provider:
    """
    Defines for each of the supported providers

    Non-Dummy drivers are sorted in alphabetical order. Please preserve this
    ordering when adding new drivers.

    :cvar DUMMY: Example provider
    :cvar ALIYUN_OSS: Aliyun OSS storage driver
    :cvar AURORAOBJECTS: AuroraObjects storage driver
    :cvar AZURE_BLOBS: Azure Blob Storage driver
    :cvar BACKBLAZE_B2: Backblaze B2 Cloud Storage driver
    :cvar CLOUDFILES: CloudFiles
    :cvar DIGITALOCEAN_SPACES: Digital Ocean Spaces driver
    :cvar GOOGLE_STORAGE Google Storage
    :cvar KTUCLOUD: KT UCloud Storage driver
    :cvar LOCAL: Local storage driver
    :cvar NIMBUS: Nimbus.io driver
    :cvar NINEFOLD: Ninefold
    :cvar OPENSTACK_SWIFT: OpenStack Swift driver
    :cvar S3: Amazon S3 US
    :cvar S3_AP_NORTHEAST: Amazon S3 Asia North East (Tokyo)
    :cvar S3_AP_NORTHEAST1: Amazon S3 Asia North East (Tokyo)
    :cvar S3_AP_NORTHEAST2: Amazon S3 Asia North East (Seoul)
    :cvar S3_AP_SOUTH: Amazon S3 Asia South (Mumbai)
    :cvar S3_AP_SOUTHEAST: Amazon S3 Asia South East (Singapore)
    :cvar S3_AP_SOUTHEAST2: Amazon S3 Asia South East 2 (Sydney)
    :cvar S3_CA_CENTRAL: Amazon S3 Canada (Central)
    :cvar S3_CN_NORTH: Amazon S3 CN North (Beijing)
    :cvar S3_EU_WEST: Amazon S3 EU West (Ireland)
    :cvar S3_EU_WEST2: Amazon S3 EU West 2 (London)
    :cvar S3_EU_CENTRAL: Amazon S3 EU Central (Frankfurt)
    :cvar S3_EU_NORTH1: Amazon S3 EU North 1 (Stockholm)
    :cvar S3_SA_EAST: Amazon S3 South America East (Sao Paulo)
    :cvar S3_US_EAST2: Amazon S3 US East 2 (Ohio)
    :cvar S3_US_WEST: Amazon S3 US West (Northern California)
    :cvar S3_US_WEST_OREGON: Amazon S3 US West 2 (Oregon)
    :cvar S3_US_GOV_WEST: Amazon S3 GovCloud (US)
    :cvar S3_RGW: S3 RGW
    :cvar S3_RGW_OUTSCALE: OUTSCALE S3 RGW
    """

    DUMMY = "dummy"
    ALIYUN_OSS = "aliyun_oss"
    AURORAOBJECTS = "auroraobjects"
    AZURE_BLOBS = "azure_blobs"
    BACKBLAZE_B2 = "backblaze_b2"
    CLOUDFILES = "cloudfiles"
    DIGITALOCEAN_SPACES = "digitalocean_spaces"
    GOOGLE_STORAGE = "google_storage"
    KTUCLOUD = "ktucloud"
    LOCAL = "local"
    NIMBUS = "nimbus"
    NINEFOLD = "ninefold"
    OPENSTACK_SWIFT = "openstack_swift"
    S3 = "s3"
    S3_AP_NORTHEAST = "s3_ap_northeast"
    S3_AP_NORTHEAST1 = "s3_ap_northeast_1"
    S3_AP_NORTHEAST2 = "s3_ap_northeast_2"
    S3_AP_SOUTH = "s3_ap_south"
    S3_AP_SOUTHEAST = "s3_ap_southeast"
    S3_AP_SOUTHEAST2 = "s3_ap_southeast2"
    S3_CA_CENTRAL = "s3_ca_central"
    S3_CN_NORTH = "s3_cn_north"
    S3_CN_NORTHWEST = "s3_cn_northwest"
    S3_EU_WEST = "s3_eu_west"
    S3_EU_WEST2 = "s3_eu_west_2"
    S3_EU_CENTRAL = "s3_eu_central"
    S3_EU_NORTH1 = "s3_eu_north_1"
    S3_SA_EAST = "s3_sa_east"
    S3_US_EAST2 = "s3_us_east_2"
    S3_US_WEST = "s3_us_west"
    S3_US_WEST_OREGON = "s3_us_west_oregon"
    S3_US_GOV_WEST = "s3_us_gov_west"
    S3_RGW = "s3_rgw"
    S3_RGW_OUTSCALE = "s3_rgw_outscale"
    MINIO = "minio"
    SCALEWAY = "scaleway"
    OVH = "ovh"

    # Deprecated
    CLOUDFILES_US = "cloudfiles_us"
    CLOUDFILES_UK = "cloudfiles_uk"
    CLOUDFILES_SWIFT = "cloudfiles_swift"


OLD_CONSTANT_TO_NEW_MAPPING = {
    # CloudFiles
    Provider.CLOUDFILES_US: Provider.CLOUDFILES,
    Provider.CLOUDFILES_UK: Provider.CLOUDFILES_UK,
    Provider.CLOUDFILES_SWIFT: Provider.OPENSTACK_SWIFT,
}


class ContainerError(LibcloudError):
    error_type = "ContainerError"

    def __init__(self, value, driver, container_name):
        self.container_name = container_name
        super().__init__(value=value, driver=driver)

    def __str__(self):
        return "<{} in {}, container={}, value={}>".format(
            self.error_type,
            repr(self.driver),
            self.container_name,
            self.value,
        )


class ObjectError(LibcloudError):
    error_type = "ContainerError"

    def __init__(self, value, driver, object_name):
        self.object_name = object_name
        super().__init__(value=value, driver=driver)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<{} in {}, value={}, object = {}>".format(
            self.error_type,
            repr(self.driver),
            self.value,
            self.object_name,
        )


class ContainerAlreadyExistsError(ContainerError):
    error_type = "ContainerAlreadyExistsError"


class ContainerDoesNotExistError(ContainerError):
    error_type = "ContainerDoesNotExistError"


class ContainerIsNotEmptyError(ContainerError):
    error_type = "ContainerIsNotEmptyError"


class ObjectDoesNotExistError(ObjectError):
    error_type = "ObjectDoesNotExistError"


class ObjectHashMismatchError(ObjectError):
    error_type = "ObjectHashMismatchError"


class InvalidContainerNameError(ContainerError):
    error_type = "InvalidContainerNameError"



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.utils.misc import dict2str, str2list, str2dicts
from libcloud.compute.base import Node
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.cloudsigma import CloudSigmaNodeDriver, CloudSigmaZrhNodeDriver


class CloudSigmaAPI10BaseTestCase:
    should_list_locations = False

    driver_klass = CloudSigmaZrhNodeDriver
    driver_kwargs = {}

    def setUp(self):
        self.driver_klass.connectionCls.conn_class = CloudSigmaHttp
        CloudSigmaHttp.type = None
        self.driver = self.driver_klass(*self.driver_args, **self.driver_kwargs)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertTrue(isinstance(nodes, list))
        self.assertEqual(len(nodes), 1)

        node = nodes[0]
        self.assertEqual(node.public_ips[0], "1.2.3.4")
        self.assertEqual(node.extra["smp"], 1)
        self.assertEqual(node.extra["cpu"], 1100)
        self.assertEqual(node.extra["mem"], 640)

    def test_list_sizes(self):
        images = self.driver.list_sizes()
        self.assertEqual(len(images), 10)

    def test_list_images(self):
        sizes = self.driver.list_images()
        self.assertEqual(len(sizes), 10)

    def test_start_node(self):
        nodes = self.driver.list_nodes()
        node = nodes[0]
        self.assertTrue(self.driver.ex_start_node(node))

    def test_shutdown_node(self):
        nodes = self.driver.list_nodes()
        node = nodes[0]
        self.assertTrue(self.driver.ex_stop_node(node))
        self.assertTrue(self.driver.ex_shutdown_node(node))

    def test_reboot_node(self):
        node = self.driver.list_nodes()[0]
        self.assertTrue(self.driver.reboot_node(node))

    def test_destroy_node(self):
        node = self.driver.list_nodes()[0]
        self.assertTrue(self.driver.destroy_node(node))
        self.driver.list_nodes()

    def test_create_node(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        node = self.driver.create_node(name="cloudsigma node", image=image, size=size)
        self.assertTrue(isinstance(node, Node))

    def test_ex_static_ip_list(self):
        ips = self.driver.ex_static_ip_list()
        self.assertEqual(len(ips), 3)

    def test_ex_static_ip_create(self):
        result = self.driver.ex_static_ip_create()
        self.assertEqual(len(result), 2)
        self.assertEqual(len(list(result[0].keys())), 6)
        self.assertEqual(len(list(result[1].keys())), 6)

    def test_ex_static_ip_destroy(self):
        result = self.driver.ex_static_ip_destroy("1.2.3.4")
        self.assertTrue(result)

    def test_ex_drives_list(self):
        result = self.driver.ex_drives_list()
        self.assertEqual(len(result), 2)

    def test_ex_drive_destroy(self):
        result = self.driver.ex_drive_destroy(
            # @@TR: this should be soft-coded:
            "d18119ce_7afa_474a_9242_e0384b160220"
        )
        self.assertTrue(result)

    def test_ex_set_node_configuration(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.ex_set_node_configuration(node, **{"smp": 2})
        self.assertTrue(result)

    def test_str2dicts(self):
        string = "mem 1024\ncpu 2200\n\nmem2048\\cpu 1100"
        result = str2dicts(string)
        self.assertEqual(len(result), 2)

    def test_str2list(self):
        string = "ip 1.2.3.4\nip 1.2.3.5\nip 1.2.3.6"
        result = str2list(string)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], "1.2.3.4")
        self.assertEqual(result[1], "1.2.3.5")
        self.assertEqual(result[2], "1.2.3.6")

    def test_dict2str(self):
        d = {"smp": 5, "cpu": 2200, "mem": 1024}
        result = dict2str(d)
        self.assertTrue(len(result) > 0)
        self.assertTrue(result.find("smp 5") >= 0)
        self.assertTrue(result.find("cpu 2200") >= 0)
        self.assertTrue(result.find("mem 1024") >= 0)


class CloudSigmaAPI10DirectTestCase(CloudSigmaAPI10BaseTestCase, unittest.TestCase):
    driver_klass = CloudSigmaZrhNodeDriver
    driver_args = ("foo", "bar")
    driver_kwargs = {}


class CloudSigmaAPI10IndiretTestCase(CloudSigmaAPI10BaseTestCase, unittest.TestCase):
    driver_klass = CloudSigmaNodeDriver
    driver_args = ("foo", "bar")
    driver_kwargs = {"api_version": "1.0"}


class CloudSigmaHttp(MockHttp):
    fixtures = ComputeFileFixtures("cloudsigma")

    def _drives_standard_info(self, method, url, body, headers):
        body = self.fixtures.load("drives_standard_info.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _servers_62fe7cde_4fb9_4c63_bd8c_e757930066a0_start(self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _servers_62fe7cde_4fb9_4c63_bd8c_e757930066a0_stop(self, method, url, body, headers):
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.OK])

    def _servers_62fe7cde_4fb9_4c63_bd8c_e757930066a0_destroy(self, method, url, body, headers):
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _drives_d18119ce_7afa_474a_9242_e0384b160220_clone(self, method, url, body, headers):
        body = self.fixtures.load("drives_clone.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _drives_a814def5_1789_49a0_bf88_7abe7bb1682a_info(self, method, url, body, headers):
        body = self.fixtures.load("drives_single_info.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _drives_info(self, method, url, body, headers):
        body = self.fixtures.load("drives_info.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _servers_create(self, method, url, body, headers):
        body = self.fixtures.load("servers_create.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _servers_info(self, method, url, body, headers):
        body = self.fixtures.load("servers_info.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _resources_ip_list(self, method, url, body, headers):
        body = self.fixtures.load("resources_ip_list.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _resources_ip_create(self, method, url, body, headers):
        body = self.fixtures.load("resources_ip_create.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _resources_ip_1_2_3_4_destroy(self, method, url, body, headers):
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.OK])

    def _drives_d18119ce_7afa_474a_9242_e0384b160220_destroy(self, method, url, body, headers):
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.OK])

    def _servers_62fe7cde_4fb9_4c63_bd8c_e757930066a0_set(self, method, url, body, headers):
        body = self.fixtures.load("servers_set.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.compute.providers import Provider
from libcloud.compute.drivers.cloudstack import CloudStackNodeDriver

__all__ = ["AuroraComputeRegion", "AuroraComputeNodeDriver"]


class AuroraComputeRegion:
    AMS = "Amsterdam"
    RTD = "Rotterdam"
    MIA = "Miami"
    LAX = "Los Angeles"
    TYO = "Tokyo"
    BCN = "Barcelona"


REGION_ENDPOINT_MAP = {
    AuroraComputeRegion.AMS: "/ams",
    AuroraComputeRegion.RTD: "/rtd",
    AuroraComputeRegion.MIA: "/mia",
    AuroraComputeRegion.LAX: "/lax",
    AuroraComputeRegion.TYO: "/tyo",
    AuroraComputeRegion.BCN: "/bcn",
}


class AuroraComputeNodeDriver(CloudStackNodeDriver):
    type = Provider.AURORACOMPUTE
    name = "PCextreme AuroraCompute"
    website = "https://www.pcextreme.com/aurora/compute"

    def __init__(self, key, secret, path=None, host=None, url=None, region=None):
        if host is None:
            host = "api.auroracompute.eu"

        if path is None:
            path = REGION_ENDPOINT_MAP.get(region, "/ams")

        super().__init__(key=key, secret=secret, host=host, path=path, secure=True)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import hashlib
from typing import List

from libcloud.http import LibcloudConnection
from libcloud.utils.py3 import httplib
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import InvalidCredsError
from libcloud.utils.connection import get_response_object

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore


__all__ = ["OvhResponse", "OvhConnection"]

API_HOST = "api.ovh.com"
API_ROOT = "/1.0"
# From https://us.ovhcloud.com/about/company/data-centers
LOCATIONS = {
    "BHS1": {"id": "BHS1", "name": "Beauharnois, Quebec 1", "country": "CA"},
    "BHS2": {"id": "BHS2", "name": "Beauharnois, Quebec 2", "country": "CA"},
    "BHS3": {"id": "BHS3", "name": "Beauharnois, Quebec 3", "country": "CA"},
    "BHS4": {"id": "BHS4", "name": "Beauharnois, Quebec 4", "country": "CA"},
    "BHS5": {"id": "BHS5", "name": "Beauharnois, Quebec 5", "country": "CA"},
    "BHS6": {"id": "BHS6", "name": "Beauharnois, Quebec 6", "country": "CA"},
    "DC1": {"id": "DC1", "name": "Paris DC1", "country": "FR"},
    "FRA1": {"id": "FRA1", "name": "Frankfurt 1", "country": "DE"},
    "GRA1": {"id": "GRA1", "name": "Gravelines 1", "country": "FR"},
    "GRA2": {"id": "GRA2", "name": "Gravelines 2", "country": "FR"},
    "GSW": {"id": "GSW", "name": "Paris GSW", "country": "FR"},
    "HIL1": {"id": "HIL1", "name": "Hillsboro, Oregon 1", "country": "US"},
    "LON1": {"id": "LON1", "name": "London 1", "country": "UK"},
    "P19": {"id": "P19", "name": "Paris P19", "country": "FR"},
    "RBX1": {"id": "RBX1", "name": "Roubaix 1", "country": "FR"},
    "RBX2": {"id": "RBX2", "name": "Roubaix 2", "country": "FR"},
    "RBX3": {"id": "RBX3", "name": "Roubaix 3", "country": "FR"},
    "RBX4": {"id": "RBX4", "name": "Roubaix 4", "country": "FR"},
    "RBX5": {"id": "RBX5", "name": "Roubaix 5", "country": "FR"},
    "RBX6": {"id": "RBX6", "name": "Roubaix 6", "country": "FR"},
    "RBX7": {"id": "RBX7", "name": "Roubaix 7", "country": "FR"},
    "SBG1": {"id": "SBG1", "name": "Strasbourg 1", "country": "FR"},
    "SBG2": {"id": "SBG2", "name": "Strasbourg 2", "country": "FR"},
    "SBG3": {"id": "SBG3", "name": "Strasbourg 3", "country": "FR"},
    "SGP1": {"id": "SGP1", "name": "Singapore 1", "country": "SG"},
    "SYD1": {"id": "SYD1", "name": "Sydney 1", "country": "AU"},
    "VIN1": {"id": "VIN1", "name": "Vint Hill, Virginia 1", "country": "US"},
    "WAW1": {"id": "WAW1", "name": "Warsaw 1", "country": "PL"},
}
DEFAULT_ACCESS_RULES = [
    {"method": "GET", "path": "/*"},
    {"method": "POST", "path": "/*"},
    {"method": "PUT", "path": "/*"},
    {"method": "DELETE", "path": "/*"},
]


class OvhException(Exception):
    pass


class OvhResponse(JsonResponse):
    def parse_error(self):
        response = super().parse_body()
        response = response or {}

        if response.get("errorCode", None) == "INVALID_SIGNATURE":
            raise InvalidCredsError(
                "Signature validation failed, probably " "using invalid credentials"
            )

        return self.body


class OvhConnection(ConnectionUserAndKey):
    """
    A connection to the Ovh API

    Wraps SSL connections to the Ovh API, automagically injecting the
    parameters that the API needs for each request.
    """

    host = API_HOST
    request_path = API_ROOT
    responseCls = OvhResponse
    timestamp = None
    ua = []  # type: List[str]
    LOCATIONS = LOCATIONS
    _timedelta = None

    allow_insecure = True

    def __init__(self, user_id, *args, **kwargs):
        region = kwargs.pop("region", "")
        if region:
            self.host = ("{}.{}".format(region, API_HOST)).lstrip(".")
        else:
            self.host = API_HOST
        self.consumer_key = kwargs.pop("ex_consumer_key", None)
        if self.consumer_key is None:
            consumer_key_json = self.request_consumer_key(user_id)
            msg = (
                "Your consumer key isn't validated, "
                "go to '%(validationUrl)s' for valid it. After instantiate "
                "your driver with \"ex_consumer_key='%(consumerKey)s'\"." % consumer_key_json
            )
            raise OvhException(msg)
        super().__init__(user_id, *args, **kwargs)

    def request_consumer_key(self, user_id):
        action = self.request_path + "/auth/credential"
        data = json.dumps({"accessRules": DEFAULT_ACCESS_RULES, "redirection": "http://ovh.com"})
        headers = {
            "Content-Type": "application/json",
            "X-Ovh-Application": user_id,
        }
        httpcon = LibcloudConnection(host=self.host, port=443)

        try:
            httpcon.request(method="POST", url=action, body=data, headers=headers)
        except Exception as e:
            handle_and_rethrow_user_friendly_invalid_region_error(host=self.host, e=e)

        response = OvhResponse(httpcon.getresponse(), httpcon)

        if response.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()

        json_response = response.parse_body()
        httpcon.close()
        return json_response

    def get_timestamp(self):
        if not self._timedelta:
            url = "https://{}{}/auth/time".format(self.host, API_ROOT)
            response = get_response_object(url=url, method="GET", headers={})
            if not response or not response.body:
                raise Exception("Failed to get current time from Ovh API")

            timestamp = int(response.body)
            self._timedelta = timestamp - int(time.time())
        return int(time.time()) + self._timedelta

    def make_signature(self, method, action, params, data, timestamp):
        full_url = "https://{}{}".format(self.host, action)
        if params:
            full_url += "?"
            for key, value in params.items():
                full_url += "{}={}&".format(key, value)
            full_url = full_url[:-1]
        sha1 = hashlib.sha1()
        base_signature = "+".join(
            [
                self.key,
                self.consumer_key,
                method.upper(),
                full_url,
                data if data else "",
                str(timestamp),
            ]
        )
        sha1.update(base_signature.encode())
        signature = "$1$" + sha1.hexdigest()
        return signature

    def add_default_params(self, params):
        return params

    def add_default_headers(self, headers):
        headers.update(
            {
                "X-Ovh-Application": self.user_id,
                "X-Ovh-Consumer": self.consumer_key,
                "Content-type": "application/json",
            }
        )
        return headers

    def request(self, action, params=None, data=None, headers=None, method="GET", raw=False):
        data = json.dumps(data) if data else None
        timestamp = self.get_timestamp()
        signature = self.make_signature(method, action, params, data, timestamp)
        headers = headers or {}
        headers.update({"X-Ovh-Timestamp": timestamp, "X-Ovh-Signature": signature})

        try:
            return super().request(
                action,
                params=params,
                data=data,
                headers=headers,
                method=method,
                raw=raw,
            )
        except Exception as e:
            handle_and_rethrow_user_friendly_invalid_region_error(host=self.host, e=e)


def handle_and_rethrow_user_friendly_invalid_region_error(host, e):
    """
    Utility method which throws a more user-friendly error in case "name or
    service not known" error is received when sending a request.

    In most cases this error indicates user passed invalid ``region`` argument
    to the driver constructor.
    """
    msg = str(e).lower()

    error_messages_to_throw = [
        "name or service not known",
        "nodename nor servname provided, or not known",
        "getaddrinfo failed",
    ]

    if any([value for value in error_messages_to_throw if value in msg]):
        raise ValueError(
            'Received "name or service not known" error '
            "when sending a request. This likely "
            "indicates invalid region argument was "
            "passed to the driver constructor."
            "Used host: %s. Original error: %s" % (host, str(e))
        )

    raise e
